@@ -1,5 +1,5 @@
 /* 
- * Copyright (c) 2006 NTT DATA CORPORATION.
+ * Copyright (c) 2006-2007 NTT DATA CORPORATION.
  * All rights reserved.
  *
  * derived from original pg_senna.c developed by Daisuke Maki
@@ -62,10 +62,10 @@
  * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
  */
 
+#include "postgres.h"
 #include <dirent.h>
 #include <sys/stat.h>
 #include <unistd.h>
-#include "postgres.h"
 #include <mb/pg_wchar.h>
 #include "access/genam.h"
 #include "access/heapam.h"
@@ -74,6 +74,7 @@
 #include "catalog/index.h"
 #include "catalog/catalog.h"
 #include "storage/smgr.h"
+#include "storage/lock.h"
 #include "commands/vacuum.h"
 #include "executor/executor.h"
 #include "miscadmin.h"
@@ -82,6 +83,18 @@
 #include "nodes/relation.h"
 #include "storage/itemptr.h"
 #include "utils/guc.h"
+#include "funcapi.h"
+#include "utils/relcache.h"
+
+#if defined(POSTGRES81) && defined(WIN32)
+extern DLLIMPORT char *DataDir; 
+extern DLLIMPORT double cpu_operator_cost;
+#endif /* defined(POSTGRES81) && defined(WIN32) */
+
+#if defined(POSTGRES82) || defined(POSTGRES83)
+#include "access/reloptions.h"
+PG_MODULE_MAGIC;
+#endif
 
 PG_FUNCTION_INFO_V1(pgs2insert);
 PG_FUNCTION_INFO_V1(pgs2beginscan);
@@ -95,77 +108,115 @@ PG_FUNCTION_INFO_V1(pgs2build);
 PG_FUNCTION_INFO_V1(pgs2buildb);
 PG_FUNCTION_INFO_V1(pgs2buildu);
 PG_FUNCTION_INFO_V1(pgs2bulkdelete);
+PG_FUNCTION_INFO_V1(pgs2vacuumcleanup);
 PG_FUNCTION_INFO_V1(pgs2costestimate);
 PG_FUNCTION_INFO_V1(pgs2contain);
 PG_FUNCTION_INFO_V1(pgs2nop);
 PG_FUNCTION_INFO_V1(pgs2getscore);
 PG_FUNCTION_INFO_V1(pgs2getscore_simple);
 PG_FUNCTION_INFO_V1(pgs2getnhits);
-PG_FUNCTION_INFO_V1(pgs2destroy);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+PG_FUNCTION_INFO_V1(pgs2options);
+#endif
+PG_FUNCTION_INFO_V1(pgs2indexcache);
 
 
-typedef struct {
-  Oid relid;
-  Oid namespace;
-  char name[NAMEDATALEN];
-  sen_index *index;
-  int using;
-  IndexScanDesc s;
-} index_info;
+inline static void *scan_stat_open(sen_records *records, index_info *ii);
+inline static void scan_stat_close(void *opaque);
 
-typedef struct {
-  FmgrInfo filterFn;
-} support_funcs;
-
-typedef struct {
-  support_funcs funcs;
-  double indtuples;
-  index_info *index;
-} build_stat;
-
-typedef struct {
-  sen_records *records;
-  index_info *index;
-  sen_recordh *mark;
-} scan_stat;
-
-static index_info index_cache[PGS2_MAX_N_INDEX_CACHE];
+static index_info *index_cache;
+static index_info *last_used_cache;
 static scan_stat *curr_scan_stat;
 static int last_nhits = 0;
+static int max_n_index_cache = MAX_N_INDEX_CACHE;
 
-inline char *
-text2cstr(text *t)
+
+static void
+pgs2_logger_func(int level, const char *time, const char *title,
+                 const char *msg, const char *location, void *func_arg)
 {
-  int32 size = VARSIZE(t) - VARHDRSZ;
-  char *c = palloc(size + 1);
-  memcpy(c, t->vl_dat, size);
-  c[size] = '\0';
-  return c;
+  const char slev[] = " EACewnid-";
+  elog(LOG, "pgsenna2: |%c|%s %s %s", *(slev + level), title, msg, location);
 }
 
-inline static void
+static sen_logger_info pgs2_sen_logger_info = {sen_log_warning,
+                                               SEN_LOG_TIME | SEN_LOG_MESSAGE,
+                                               pgs2_logger_func};
+
+void __attribute__ ((constructor)) pgs2_init(void) {
+  sen_rc rc;
+  if (sen_logger_info_set(&pgs2_sen_logger_info)) {
+    elog(INFO, "pgsenna2: cannot logger_info_set.");
+  } else {
+    elog(DEBUG1, "pgsenna2: logger_info_set.");
+  }
+  rc = sen_init();
+  if (rc != sen_success) {
+    elog(ERROR, "pgsenna2: sen_init() failed %d", rc);
+  }
+};
+
+void
 sen_check_init(void)
 {
   static int initialized = 0;
   int i;
-  sen_rc rc;
   index_info *ii;
+  const char *option;
+
   if (initialized) { return ; }
-  if ((rc = sen_init())) { elog(ERROR, "pgsenna2: sen_init() failed %d", rc); }
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
+#ifndef __GNUC__
+  pgs2_init();
+#endif /* __GNUC__ */
+  option = GetConfigOption("ludia.max_n_index_cache");
+  if (option) {
+    max_n_index_cache = atoi(option);
+    if(max_n_index_cache <= 0) {
+      elog(ERROR, "pgsenna2: value of max_n_index_cache is invalid: %d", max_n_index_cache);
+    }
+  } else {
+    elog(DEBUG1, "pgsenna2: value of max_n_index_cache = %d", max_n_index_cache);
+  }
+  index_cache = (index_info*)malloc(max_n_index_cache * sizeof(*index_cache));
+  last_used_cache = NULL;
+  for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
     ii->namespace = 0;
     ii->relid = 0;
     ii->name[0] = '\0';
     ii->index = NULL;
+    ii->using = 0;
     ii->s = NULL;
+    ii->previous = NULL;
+    ii->next = NULL;
   }
   initialized++;
 }
 
+
+inline static void
+index_list_update(index_info *ii) {
+  if (last_used_cache == ii) {
+    return;
+  }
+  if (ii->previous) {
+    ((index_info *)ii->previous)->next = ii->next;
+  }
+  if (ii->next) {
+    ((index_info *)ii->next)->previous = ii->previous;
+  }
+  if (last_used_cache) {
+    last_used_cache->previous = ii;
+  }
+  ii->next = last_used_cache;
+  ii->previous = NULL;
+  last_used_cache = ii;
+}
+
+
 /* todo: SizeOfIptrData may be more suitable.  */
 #define KEY_SIZE (sizeof(ItemPointerData))
 
-inline static index_info *
+index_info *
 index_info_open(Relation r, int createp, int flags)
 {
   int i;
@@ -175,25 +226,46 @@ index_info_open(Relation r, int createp, int flags)
   Oid relid = RelationGetRelid(r);
   char *name = RelationGetRelationName(r);
   sen_check_init();
-  cand = NULL;
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
-    if (ii->index && 
-        namespace == ii->namespace &&
-        relid == ii->relid && 
-        !strcmp(ii->name, name)) {
-      if (createp) { /* for REINDEX (and CLUSTER) */
-	cand = ii;
-	break;
+  ii = last_used_cache;
+  for (i = max_n_index_cache - 1; i; i--) {
+    if(!ii) {
+      break;
+    } else {
+      if (ii->index && 
+          namespace == ii->namespace &&
+          relid == ii->relid && 
+          !strcmp(ii->name, name)) {
+        if (createp) { /* for REINDEX (and CLUSTER) */
+          break;
+        } else {
+          index_list_update(ii);
+          return ii;
+        }
       } else {
-	return ii;
+        ii = ii->next;
+      }
+    }
+  }
+  cand = ii;  // cand is oldest or NULL
+  if (!cand) {
+    for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
+      if (!ii->using) {
+        cand = ii;
+        break;
       }
     }
-    if (!ii->using) { cand = ii; }
   }
   if (!cand) {
-    elog(ERROR, "pgsenna2: n of indices in use reached max(%d)", PGS2_MAX_N_INDEX_CACHE);
+    elog(ERROR, "pgsenna2: n of indices in use reached max(%d)", max_n_index_cache);
   }
+  index_list_update(cand);
+
   sen_index_close(cand->index);
+  if(cand->s && cand->s->opaque) {
+    scan_stat_close(cand->s->opaque);
+    cand->s->opaque = NULL;
+    cand->s = NULL;
+  }
   cand->namespace = namespace;
   cand->relid = relid;
   strcpy(cand->name, name);
@@ -206,7 +278,9 @@ index_info_open(Relation r, int createp, int flags)
     pfree(pathname);
   }
   if (createp) {
+    int initial_n_segments = INITIAL_N_SEGMENTS;
     sen_encoding encoding;
+    const char *option;
     switch (GetDatabaseEncoding()) {
     case PG_UTF8:
       encoding = sen_enc_utf8;
@@ -221,9 +295,18 @@ index_info_open(Relation r, int createp, int flags)
       encoding = sen_enc_default;
     }
     sen_index_remove(path);
+    option = GetConfigOption("ludia.initial_n_segments");
+    if (option) {
+      initial_n_segments = atoi(option);
+      if(initial_n_segments <= 0) {
+        elog(ERROR, "pgsenna2: value of initial_n_segments is invalid: %d", initial_n_segments);
+      }
+    } else {
+      elog(DEBUG1, "pgsenna2: value of initial_n_segments = %d", initial_n_segments);
+    }
     cand->index = sen_index_create(path,
                                    KEY_SIZE,
-                                   flags, 0, encoding);
+                                   flags, initial_n_segments, encoding);
     if (!cand->index) {
       elog(ERROR, "pgsenna2: index create failed (%s)", path);
     }
@@ -237,12 +320,14 @@ index_info_open(Relation r, int createp, int flags)
   return cand;
 }
 
-inline static void
+
+void
 index_info_close(index_info *ii)
 {
   /* ii->using = 0; */
 }
 
+
 inline static void *
 scan_stat_open(sen_records *records, index_info *ii)
 {
@@ -255,6 +340,7 @@ scan_stat_open(sen_records *records, index_info *ii)
   return (void *) ss;
 }
 
+
 inline static void
 scan_stat_close(void *opaque)
 {
@@ -267,6 +353,25 @@ scan_stat_close(void *opaque)
   curr_scan_stat = NULL;
 }
 
+
+inline static void
+scan_stat_close_all()
+{
+  int i = 0;
+  index_info *ii;
+  for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
+    if (!ii->using) {
+      return;
+    }
+    if(ii->s && ii->s->opaque) {
+      scan_stat_close(ii->s->opaque);
+      ii->s->opaque = NULL;
+      ii->s = NULL;
+    }
+  }
+}
+
+
 inline static void
 init_support_funcs(support_funcs *funcs, Relation index)
 {
@@ -275,6 +380,7 @@ init_support_funcs(support_funcs *funcs, Relation index)
                  CurrentMemoryContext);
 }
 
+
 inline static void
 do_insert(Relation r, ItemPointer tid, index_info *ii,
           Datum *values, bool *isnull, support_funcs *funcs)
@@ -290,17 +396,40 @@ do_insert(Relation r, ItemPointer tid, index_info *ii,
       t = DatumGetTextP(value);
       c = text2cstr(t);
       if (c && *c) {
-        rc = sen_index_upd(ii->index, (void *)tid, NULL, c);
+        LOCKTAG locktag;
+        LockAcquireResult res = LOCKACQUIRE_OK;
+
+        SET_LOCKTAG_RELATION(locktag, MyDatabaseId, ii->relid);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+        res = LockAcquire(&locktag, ShareUpdateExclusiveLock, 0, 0);
+#else
+        res = LockAcquire(DEFAULT_LOCKMETHOD, &locktag, r->rd_istemp,
+                          ShareUpdateExclusiveLock, 0, 0);
+#endif
+        if(res == LOCKACQUIRE_OK) {
+          rc = sen_index_upd(ii->index, (void *)tid, NULL, 0, c, strlen(c));
+        } else {
+          elog(ERROR, "pgsenna2: cannot LockAcquire while do_insert (%d)", res);
+        }
+        if (rc != sen_success) {
+          elog(ERROR, "pgsenna2: sen_index_upd failed while do_insert (%d)", rc);
+        }
+#if defined(POSTGRES82) || defined(POSTGRES83)
+        LockRelease(&locktag, ShareUpdateExclusiveLock, 0);        
+#else
+        LockRelease(DEFAULT_LOCKMETHOD, &locktag,
+                    ShareUpdateExclusiveLock, 0);
+#endif
       } else {
         if (!(sen_sym_get(ii->index->keys, (void *)tid))) { rc = sen_other_error; }
       }
       pfree(c);
       if (t != (text *) value) {
-        sen_log("pgsenna2: palloced when untoasted (%p)", t);
+        elog(DEBUG1, "pgsenna2: palloced when untoasted (%p)", t);
         pfree(t);
       }
       if (value != values[i]) {
-        sen_log("pgsenna2: palloced in filter function (%p)", value);
+        elog(DEBUG1, "pgsenna2: palloced in filter function (%p)", DatumGetPointer(value));
         pfree((text *)value);
       }
       if (rc) {
@@ -310,6 +439,7 @@ do_insert(Relation r, ItemPointer tid, index_info *ii,
   }
 }
 
+
 static void
 buildCallback(Relation index,
               HeapTuple htup,
@@ -323,12 +453,16 @@ buildCallback(Relation index,
   bs->indtuples += 1;
 }
 
+
 Datum
 pgs2build(PG_FUNCTION_ARGS)
 {
   Relation  heap = (Relation) PG_GETARG_POINTER(0);
   Relation  index = (Relation) PG_GETARG_POINTER(1);
   IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  IndexBuildResult *result;
+#endif
   build_stat bs;
   double    reltuples;
   init_support_funcs(&bs.funcs, index);
@@ -336,17 +470,29 @@ pgs2build(PG_FUNCTION_ARGS)
   bs.index = index_info_open(index, 1, SEN_INDEX_NORMALIZE);
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
   IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
   index_info_close(bs.index);
   PG_RETURN_VOID();
+#endif
 }
 
+
 Datum
 pgs2buildb(PG_FUNCTION_ARGS)
 {
   Relation  heap = (Relation) PG_GETARG_POINTER(0);
   Relation  index = (Relation) PG_GETARG_POINTER(1);
   IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  IndexBuildResult *result;
+#endif
   build_stat bs;
   double    reltuples;
   init_support_funcs(&bs.funcs, index);
@@ -354,24 +500,36 @@ pgs2buildb(PG_FUNCTION_ARGS)
   bs.index = index_info_open(index, 1, SEN_INDEX_NGRAM|SEN_INDEX_NORMALIZE);
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
   IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
   index_info_close(bs.index);
   PG_RETURN_VOID();
+#endif
 }
 
+
 Datum
 pgs2buildu(PG_FUNCTION_ARGS)
 {
   Relation  heap = (Relation) PG_GETARG_POINTER(0);
   Relation  index = (Relation) PG_GETARG_POINTER(1);
   IndexInfo  *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  IndexBuildResult *result;
+#endif
   build_stat bs;
   double    reltuples;
-  init_support_funcs(&bs.funcs, index);
-  bs.indtuples = 0;
   const char *option;
   int flags;
 
+  init_support_funcs(&bs.funcs, index);
+  bs.indtuples = 0;
   option = GetConfigOption("ludia.sen_index_flags");
   if (option == NULL) {
     elog(ERROR, "pgsenna2: value of sen_index_flags is invalid: NULL.");
@@ -380,14 +538,24 @@ pgs2buildu(PG_FUNCTION_ARGS)
   if (flags < 0) {
     elog(ERROR, "pgsenna2: value of sen_index_flags is invalid: %d.", flags);
   }
+  elog(DEBUG1, "pgsenna2: value of sen_index_flags = %d", flags);
   bs.index = index_info_open(index, 1, flags);
   reltuples = IndexBuildHeapScan(heap, index, indexInfo,
                                  buildCallback, (void *) &bs);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
+  result->heap_tuples = reltuples;
+  result->index_tuples = bs.indtuples;
+  index_info_close(bs.index);
+  PG_RETURN_POINTER(result);
+#else
   IndexCloseAndUpdateStats(heap, reltuples, index, bs.indtuples);
   index_info_close(bs.index);
   PG_RETURN_VOID();
+#endif
 }
 
+
 Datum
 pgs2insert(PG_FUNCTION_ARGS)
 {
@@ -409,36 +577,62 @@ pgs2insert(PG_FUNCTION_ARGS)
   PG_RETURN_BOOL(true);
 }
 
+
 Datum
 pgs2bulkdelete(PG_FUNCTION_ARGS)
 {
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0);
+  //IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
+  IndexBulkDeleteCallback callback = (IndexBulkDeleteCallback) PG_GETARG_POINTER(2);
+  void *callback_state = (void *) PG_GETARG_POINTER(3);
+  Relation r = info->index;
+#else
   Relation r = (Relation) PG_GETARG_POINTER(0);
   IndexBulkDeleteCallback callback = (IndexBulkDeleteCallback) PG_GETARG_POINTER(1);
   void *callback_state = (void *) PG_GETARG_POINTER(2);
+#endif
   double tuples_removed = 0;
   IndexBulkDeleteResult *result;
   index_info *ii = index_info_open(r, 0, 0);
-  sen_rc rc;
+
   sen_id id;
   ItemPointerData tid;
-  unsigned int i, nrecords = sen_sym_size(ii->index->keys);
-  sen_log("pgs2bulkdelete n=%d", nrecords);
-  for (id = 1, i = 0; i < nrecords; id++) {
-    if (!sen_sym_key(ii->index->keys, id, (void *) &tid, KEY_SIZE)) {
+  unsigned int nrecords = sen_sym_size(ii->index->keys);
+  Relation hrel = heap_open(r->rd_index->indrelid, NoLock);
+  HeapTupleData htup;
+  Buffer hbuffer;
+  IndexInfo* indexInfo = BuildIndexInfo(r);
+  EState *estate = CreateExecutorState();
+  TupleTableSlot *slot = MakeSingleTupleTableSlot(RelationGetDescr(hrel));
+
+  elog(DEBUG1, "pgsenna2: pgs2bulkdelete n=%d", nrecords);
+  
+  if (indexInfo->ii_Expressions != NIL &&
+      indexInfo->ii_ExpressionsState == NIL) {
+    indexInfo->ii_ExpressionsState = (List *)ExecPrepareExpr((Expr *)indexInfo->ii_Expressions, estate);
+  }
+
+  for (id = 0;;) {
+    if ((id = sen_sym_next(ii->index->keys, id)) == SEN_SYM_NIL) {
       break;
     }
-    if (sen_sym_at(ii->index->keys, (void *) &tid) == id) {
-      if (callback(&tid, callback_state)) {
-        rc = sen_index_del(ii->index, &tid);
-        if (rc) { sen_log("sen_index_del failed(%d)", rc); }
-	/*
-        rc = sen_sym_del(ii->index->keys, &tid);
-        if (rc) { sen_log("sen_sym_del failed(%d)", rc); }
-	*/
-        tuples_removed += 1;
+    if (!sen_sym_key(ii->index->keys, id, (void *) &tid, KEY_SIZE)) {
+      elog(ERROR, "pgsenna2: sen_sym_key failed.");
+    }
+    if (callback(&tid, callback_state)) {
+      htup.t_self = tid;
+      if (heap_fetch(hrel, SnapshotAny, &htup, &hbuffer, false, NULL)) {
+        (GetPerTupleExprContext(estate))->ecxt_scantuple = slot;
+        update_index_with_tuple(indexInfo, ii, estate, &htup, NULL);
       }
+      ReleaseBuffer(hbuffer);
     }
   }
+  FreeExecutorState(estate);
+  ExecDropSingleTupleTableSlot(slot);
+  heap_close(hrel, NoLock);
+
   result = (IndexBulkDeleteResult *) palloc0(sizeof(IndexBulkDeleteResult));
   result->num_pages = 1;
   result->num_index_tuples = sen_sym_size(ii->index->keys);
@@ -446,6 +640,19 @@ pgs2bulkdelete(PG_FUNCTION_ARGS)
   PG_RETURN_POINTER(result);
 }
 
+
+Datum
+pgs2vacuumcleanup(PG_FUNCTION_ARGS)
+{
+  /* IndexVacuumInfo *info = (IndexVacuumInfo *) PG_GETARG_POINTER(0); */
+  IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
+
+  /* under construction... */
+  
+  PG_RETURN_POINTER(stats);
+}
+
+
 Datum
 pgs2beginscan(PG_FUNCTION_ARGS)
 {
@@ -457,35 +664,45 @@ pgs2beginscan(PG_FUNCTION_ARGS)
   PG_RETURN_POINTER(s);
 }
 
+
 Datum
 pgs2gettuple(PG_FUNCTION_ARGS)
 {
   IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
   ScanDirection dir = (ScanDirection) PG_GETARG_INT32(1);
   scan_stat *ss = s->opaque;
-  sen_rc rc;
+  sen_rc rc = sen_success;
   int res;
 
-  if (dir != 1) { sen_log("dir(%d) assigned in pgs2gettuple()", dir); }
+  if (dir != 1) {
+    elog(WARNING, "pgsenna2: dir(%d) assigned in pgs2gettuple()", dir);
+  }
 
-  if (!ss || !ss->records) {
+  if (!ss) {
+    scan_stat_close_all();
+    elog(ERROR, "pgsenna2: inconsistent scan, or max_n_index_cache is too small");
+  }
+  if (!ss->records) {
     elog(ERROR, "pgsenna2: inconsistent scan");
   }
 
   if (s->kill_prior_tuple) {
+#ifndef POSTGRES83
     rc = sen_index_del(ss->index->index, &(s->currentItemData));
-    if (rc) { sen_log("sen_index_del failed(%d)", rc); }
-    /*
-    rc = sen_sym_del(ss->index->index->keys, &(s->currentItemData));
-    if (rc) { sen_log("sen_sym_del failed(%d)", rc); }
-    */
+#endif
+    if (rc) {
+      elog(WARNING, "pgsenna2: sen_index_del failed(%d)", rc);
+    }
   }
 
   res = sen_records_next(ss->records, &s->xs_ctup.t_self, KEY_SIZE, NULL);
+#ifndef POSTGRES83
   if (res) { memcpy(&s->currentItemData, &s->xs_ctup.t_self, KEY_SIZE); }
+#endif
   PG_RETURN_BOOL(res ? true : false);
 }
 
+
 Datum
 pgs2getmulti(PG_FUNCTION_ARGS)
 {
@@ -496,7 +713,10 @@ pgs2getmulti(PG_FUNCTION_ARGS)
   int32 i;
   scan_stat *ss = s->opaque;
   bool more = true;
-  if (!ss || !ss->records) {
+  if (!ss) {
+    elog(ERROR, "pgsenna2: inconsistent scan, or max_n_index_cache is too small");
+  }
+  if (!ss->records) {
     elog(ERROR, "pgsenna2: inconsistent scan");
   }
   for (i = 0; i < max_tids; i++) {
@@ -504,12 +724,15 @@ pgs2getmulti(PG_FUNCTION_ARGS)
       more = false;
       break;
     }
+#ifndef POSTGRES83
     memcpy(&s->currentItemData, &tids[i], KEY_SIZE);
+#endif
   }
   *returned_tids = i;
   PG_RETURN_BOOL(more);
 }
 
+
 Datum
 pgs2rescan(PG_FUNCTION_ARGS)
 {
@@ -523,22 +746,24 @@ pgs2rescan(PG_FUNCTION_ARGS)
   sen_encoding encoding;
   sen_sel_operator op;
   const char *option;
-  int max_n_sort_result = 10000;
+  int max_n_sort_result = MAX_N_SORT_RESULT;
 
   if (!key) { elog(ERROR, "pgsenna2: access method does not support scan without scankey."); }
 
+#ifndef POSTGRES83
   ItemPointerSetInvalid(&s->currentItemData);
   ItemPointerSetInvalid(&s->currentMarkData);
+#endif
 
   scan_stat_close(s->opaque);
 
   ii = index_info_open(s->indexRelation, 0, 0);
-  rc = sen_index_info(ii->index, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &encoding);
+  rc = sen_index_info(ii->index, NULL, NULL, NULL, &encoding, NULL, NULL, NULL, NULL, NULL, NULL);
   if (rc) {
     elog(ERROR, "pgsenna2: index_info failed(%d)", rc);
   }
 
-  sen_log("nok=%d ign=%d", s->numberOfKeys, s->ignore_killed_tuples);
+  elog(DEBUG1, "pgsenna2: nok=%d ign=%d", s->numberOfKeys, s->ignore_killed_tuples);
 
   if (key && s->numberOfKeys > 0) {
     memmove(s->keyData, key, s->numberOfKeys * sizeof(ScanKeyData));
@@ -552,15 +777,21 @@ pgs2rescan(PG_FUNCTION_ARGS)
 
   for (i = 0, op = sen_sel_or; i < s->numberOfKeys; i++, op = sen_sel_and) {
     char *c;
-    sen_log("sk_flags=%d,atn=%d,strategy=%d,subtype=%d,argument=%x",
-            s->keyData[i].sk_flags,
-            s->keyData[i].sk_attno,
-            s->keyData[i].sk_strategy,
-            s->keyData[i].sk_subtype,
-            s->keyData[i].sk_argument);
+    elog(DEBUG1,
+         "pgsenna2: sk_flags=%d,atn=%d,strategy=%d,subtype=%d,argument=%p",
+         s->keyData[i].sk_flags,
+         s->keyData[i].sk_attno,
+         s->keyData[i].sk_strategy,
+         s->keyData[i].sk_subtype,
+         DatumGetPointer(s->keyData[i].sk_argument));
     if(s->keyData[i].sk_argument != 0) {
       c = text2cstr(DatumGetTextP(s->keyData[i].sk_argument));
-      if ((q = sen_query_open(c, sen_sel_or, PGS2_MAX_N_EXPRS, encoding))) {
+      if ((q = sen_query_open(c, strlen(c), sen_sel_or,
+                              PGS2_MAX_N_EXPRS, encoding))) {
+        if (sen_query_rest(q, NULL)) {
+          elog(WARNING, "pgsenna2: too many expressions (%d)",
+               sen_query_rest(q, NULL));
+        }
         rc = sen_query_exec(ii->index, q, r, op);
         sen_query_close(q);
       }
@@ -573,58 +804,78 @@ pgs2rescan(PG_FUNCTION_ARGS)
   option = GetConfigOption("ludia.max_n_sort_result");
   if (option) {
     max_n_sort_result = atoi(option);
-    if (max_n_sort_result <= 0) {
+    if (max_n_sort_result < -1) {
       elog(ERROR, "pgsenna2: value of max_n_sort_result is invalid: %d", max_n_sort_result);
     }
   } else {
-      elog(WARNING, "pgsenna2: value of max_n_sort_result = %d", max_n_sort_result);
+      elog(DEBUG1, "pgsenna2: value of max_n_sort_result = %d", max_n_sort_result);
+  }
+  if (max_n_sort_result >= 0 && sen_records_nhits(r) > 0) {
+    sen_records_sort(r, max_n_sort_result, NULL);
   }
-  sen_records_sort(r, max_n_sort_result, NULL);
   s->opaque = scan_stat_open(r, ii);
   ii->s = s;
   PG_RETURN_VOID();
 }
 
+
 Datum
 pgs2endscan(PG_FUNCTION_ARGS)
 {
   IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
   scan_stat *ss = (scan_stat *) s->opaque;
-  ss->index->s = NULL;
+  if (ss != NULL && ss->index != NULL) {
+    ss->index->s = NULL;
+  }
   scan_stat_close(ss);
   PG_RETURN_VOID();
 }
 
+
 Datum
 pgs2markpos(PG_FUNCTION_ARGS)
 {
   IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
   scan_stat *ss = (scan_stat *) s->opaque;
   ss->mark = ss->records->curr_rec;
+#ifndef POSTGRES83
   s->currentMarkData = s->currentItemData;
+#endif
   PG_RETURN_VOID();
 }
 
+
 Datum
 pgs2restrpos(PG_FUNCTION_ARGS)
 {
   IndexScanDesc s = (IndexScanDesc) PG_GETARG_POINTER(0);
   scan_stat *ss = (scan_stat *) s->opaque;
   ss->records->curr_rec = ss->mark;
+#ifndef POSTGRES83
   s->currentItemData = s->currentMarkData;
+#endif
   PG_RETURN_VOID();
 }
 
+
 Datum 
 pgs2costestimate(PG_FUNCTION_ARGS)
 {
   PlannerInfo *root = (PlannerInfo *) PG_GETARG_POINTER(0);
   IndexOptInfo *index = (IndexOptInfo *) PG_GETARG_POINTER(1);
   List *indexQuals = (List *) PG_GETARG_POINTER(2);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  //  RelOptInfo *outer_rel = (RelOptInfo *) PG_GETARG_POINTER(3);
+  Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(4);
+  Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(5);
+  Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(6);
+  double *indexCorrelation = (double *) PG_GETARG_POINTER(7);
+#else
   Cost *indexStartupCost = (Cost *) PG_GETARG_POINTER(3);
   Cost *indexTotalCost = (Cost *) PG_GETARG_POINTER(4);
   Selectivity *indexSelectivity = (Selectivity *) PG_GETARG_POINTER(5);
   double *indexCorrelation = (double *) PG_GETARG_POINTER(6);
+#endif
   double numIndexTuples = 0.0;
   double numIndexPages = 0.0;
   QualCost index_qual_cost;
@@ -661,7 +912,11 @@ pgs2costestimate(PG_FUNCTION_ARGS)
 
   *indexTotalCost = numIndexPages;
 
+#ifdef POSTGRES83
+  cost_qual_eval(&index_qual_cost, indexQuals, NULL);
+#else
   cost_qual_eval(&index_qual_cost, indexQuals);
+#endif
   qual_op_cost = cpu_operator_cost * list_length(indexQuals);
   qual_arg_cost = index_qual_cost.startup +
     index_qual_cost.per_tuple - qual_op_cost;
@@ -670,40 +925,96 @@ pgs2costestimate(PG_FUNCTION_ARGS)
   *indexStartupCost = qual_arg_cost;
   *indexTotalCost += qual_arg_cost;
   *indexTotalCost += numIndexTuples * (cpu_index_tuple_cost + qual_op_cost);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+  *indexTotalCost += -DEFAULT_RANDOM_PAGE_COST;
+#endif
   *indexCorrelation = 1.0;
-  sen_log("cost=(%f,%f,%f)", *indexStartupCost, *indexTotalCost, *indexSelectivity);
+  elog(DEBUG1, "pgsenna2: cost=(%f,%f,%f)",
+       *indexStartupCost, *indexTotalCost, *indexSelectivity);
   PG_RETURN_VOID();
 }
 
+
 Datum 
 pgs2contain(PG_FUNCTION_ARGS)
 {
   text *t = PG_GETARG_TEXT_P(0);
   text *q = PG_GETARG_TEXT_P(1);
   const char *option;
+  char enable_seqscan[16];
 
   option = GetConfigOption("ludia.enable_seqscan");
   if (option) {
-    if (!strcmp(option, "on")) {
-      char *ct = text2cstr(t);
-      char *cq = text2cstr(q);
-      bool res = strstr(ct, cq) ? true : false;
-      pfree(ct);
-      pfree(cq);
-      PG_RETURN_BOOL(res);
+    strncpy(enable_seqscan, option, 16);
+  } else {
+    strncpy(enable_seqscan, ENABLE_SEQSCAN, 16);
+  }
+  if (!strcmp(enable_seqscan, "on")) {
+    char *ct = text2cstr(t);
+    char *cq = text2cstr(q);
+    bool res = false;
+    sen_encoding encoding;
+    sen_query *sq;
+
+    switch (GetDatabaseEncoding()) {
+    case PG_UTF8:
+      encoding = sen_enc_utf8;
+      break;
+    case PG_EUC_JP:
+      encoding = sen_enc_euc_jp;
+      break;
+    case PG_SJIS:
+      encoding = sen_enc_sjis;
+      break;
+    default:
+      encoding = sen_enc_default;
     }
+    if ((sq = sen_query_open(cq, strlen(cq), sen_sel_or,
+                             PGS2_MAX_N_EXPRS, encoding))) {
+      int found, score, seqscan_flags = SEQSCAN_FLAGS;
+      unsigned int ct_strlen;
+      const char *option;
+      sen_rc rc;
+
+      if (sen_query_rest(sq, NULL)) {
+        elog(WARNING, "pgsenna2: too many expressions (%d)",
+        sen_query_rest(sq, NULL));
+      }
+      option = GetConfigOption("ludia.seqscan_flags");
+      if (option) {
+        seqscan_flags = atoi(option);
+      }
+      if (seqscan_flags < 0 || 1 < seqscan_flags) {
+        elog(ERROR, "pgsenna2: value of seqscan_flags is invalid: %d",
+             seqscan_flags);
+      }
+      ct_strlen = strlen(ct);
+      rc = sen_query_scan(sq, (const char **)&ct, &ct_strlen, 1,
+                          seqscan_flags, &found, &score);
+      if (rc != sen_success) {
+        elog(ERROR, "pgsenna2: sen_query_scan failed (%d)", rc);
+      }
+      res = (bool) found;
+      sen_query_close(sq);
+    }
+    pfree(ct);
+    pfree(cq);
+    PG_RETURN_BOOL(res);
+  } else {
     ereport(ERROR, (errmsg("pgsenna2: sequencial scan disabled.")));
   }
   ereport(ERROR, (errmsg("pgsenna2: value of enable_seqscan is invalid.")));
   PG_RETURN_BOOL(false);
 }
 
+
 Datum 
 pgs2nop(PG_FUNCTION_ARGS)
 {
   PG_RETURN_POINTER(PG_GETARG_POINTER(0));
 }
 
+
 Datum
 pgs2getscore(PG_FUNCTION_ARGS)
 {
@@ -713,7 +1024,7 @@ pgs2getscore(PG_FUNCTION_ARGS)
   ItemPointer tid = (ItemPointer)PG_GETARG_POINTER(0);
   text *t = PG_GETARG_TEXT_P(1);
   char *indexname = text2cstr(t);
-  for (i = PGS2_MAX_N_INDEX_CACHE, ii = index_cache; i; i--, ii++) {
+  for (i = max_n_index_cache, ii = index_cache; i; i--, ii++) {
     if (ii->index && !strcmp(ii->name, indexname)) { break; }
   }
   if (!i || !ii->s || !(ss = ii->s->opaque) || !ss->records) {
@@ -724,6 +1035,7 @@ pgs2getscore(PG_FUNCTION_ARGS)
   PG_RETURN_INT32(result);
 }
 
+
 Datum
 pgs2getscore_simple(PG_FUNCTION_ARGS)
 {
@@ -735,67 +1047,229 @@ pgs2getscore_simple(PG_FUNCTION_ARGS)
   PG_RETURN_INT32(result);
 }
 
+
 Datum
 pgs2getnhits(PG_FUNCTION_ARGS)
 {
   PG_RETURN_INT32(last_nhits);
 }
 
-static void
-do_dir(const char *dirname, int *n)
+
+#if defined(POSTGRES82) || defined(POSTGRES83)
+Datum
+pgs2options(PG_FUNCTION_ARGS)
 {
-  DIR *dir = opendir(dirname);
-  if (dir) {
-    int fd = dirfd(dir);
-    struct dirent *ent;
-    fchdir(fd);
-    while ((ent = readdir(dir))) {
-      const char *name = ent->d_name;
-      if (!strcmp(name, ".") || !strcmp(name, "..")) {
-        continue;
-      }
-      switch (ent->d_type) {
-      case DT_REG :
-        {
-          int len = strlen(name) - 4;
-          if (len > 0 &&
-              name[len] == '.' &&
-              name[len + 1] == 'S' &&
-              name[len + 2] == 'E' &&
-              name[len + 3] == 'N') {
-            struct stat s;
-            char *idxname = strdup(name);
-            idxname[len] = '\0';
-            if (stat(idxname, &s) == -1 && errno == ENOENT) {
-              sen_index_remove(idxname);
-              (*n)++;
-            }
-          }
-        }
-        break;
-      case DT_DIR :
-      case DT_LNK :
-        do_dir(ent->d_name, n);
-        fchdir(fd);
-        break;
-      default :
+  elog(ERROR, "pgsenna2: pgs2options is called");
+  /*
+  Datum      reloptions = PG_GETARG_DATUM(0);
+  bool       validate = PG_GETARG_BOOL(1);
+  bytea      *result;
+
+  result = default_reloptions(reloptions, validate,
+                              PGS2_MIN_FILLFACTOR,
+                              PGS2_DEFAULT_FILLFACTOR);
+  if (result)
+    PG_RETURN_BYTEA_P(result);
+  */
+  PG_RETURN_NULL();
+}
+#endif
+
+
+Datum
+pgs2indexcache(PG_FUNCTION_ARGS)
+{
+  int n = 0;
+
+  FuncCallContext     *funcctx;
+  int                  call_cntr;
+  int                  max_calls;
+  TupleDesc            tupdesc;
+  AttInMetadata       *attinmeta;
+  index_info **last_used_cache_array;
+  
+  if (SRF_IS_FIRSTCALL()) {
+    index_info *tmp_used_cache;
+    MemoryContext   oldcontext;
+    funcctx = SRF_FIRSTCALL_INIT();
+    oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
+    if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+      ereport(ERROR,
+              (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+               errmsg("pgsenna2: function returning record called in context "
+                      "that cannot accept type record")));
+    attinmeta = TupleDescGetAttInMetadata(tupdesc);
+    funcctx->attinmeta = attinmeta;
+
+    last_used_cache_array = (index_info **)palloc(max_n_index_cache * sizeof(index_info *));
+    tmp_used_cache = last_used_cache;
+    while(tmp_used_cache) {
+      last_used_cache_array[n] = tmp_used_cache;
+      tmp_used_cache = tmp_used_cache->next;
+      n++;
+      if(n > max_n_index_cache) {
+        elog(WARNING, "pgsenna2: warning happened at index cache");
         break;
       }
     }
-    closedir(dir);
+    funcctx->user_fctx = (void *)last_used_cache_array;
+    funcctx->max_calls = n;
+    MemoryContextSwitchTo(oldcontext);
+  }
+
+  funcctx = SRF_PERCALL_SETUP();
+
+  last_used_cache_array = (index_info **) funcctx->user_fctx;
+  
+  call_cntr = funcctx->call_cntr;
+  max_calls = funcctx->max_calls;
+  attinmeta = funcctx->attinmeta;
+ 
+  if (call_cntr < max_calls) {
+    int i;
+    char       **values;
+    HeapTuple    tuple;
+    Datum        result;
+    
+    values = (char **) palloc(12 * sizeof(char *));
+    for (i = 0; i < 6; i++) {
+      values[i] = (char *) palloc(64 * sizeof(char));
+    }
+    last_used_cache_array = (index_info **) funcctx->user_fctx;
+    snprintf(values[0], 64, "%d", last_used_cache_array[call_cntr]->relid);
+    snprintf(values[1], 64, "%d", last_used_cache_array[call_cntr]->namespace);
+    snprintf(values[2], 64, "%s", last_used_cache_array[call_cntr]->name);
+    snprintf(values[3], 64, "%p", last_used_cache_array[call_cntr]->previous);
+    snprintf(values[4], 64, "%p", last_used_cache_array[call_cntr]);
+    snprintf(values[5], 64, "%p", last_used_cache_array[call_cntr]->next);
+    tuple = BuildTupleFromCStrings(attinmeta, values);
+    result = HeapTupleGetDatum(tuple);
+    for (i = 0; i < 6; i++) {
+      pfree(values[i]);
+    }
+    pfree(values);
+
+    SRF_RETURN_NEXT(funcctx, result);
+  } else {
+    SRF_RETURN_DONE(funcctx);
   }
 }
 
-Datum
-pgs2destroy(PG_FUNCTION_ARGS)
+
+char* get_tuple_text(ExprContext *econtext,
+                     HeapTuple tup,
+                     int attnum,
+                     ExprState *estate)
 {
-  int n = 0;
-  DIR *dir = opendir(".");
-  if (dir) {
-    int fd = dirfd(dir);
-    do_dir(DataDir, &n);
-    fchdir(fd);
-    closedir(dir);
-  }
-  PG_RETURN_INT32(n);
+  Datum iDatum;
+  text *t;
+  char *c;
+  bool isNull;
+  ExecStoreTuple(tup, econtext->ecxt_scantuple, InvalidBuffer, false);
+  if (!estate)
+    iDatum = slot_getattr(econtext->ecxt_scantuple, attnum, &isNull);
+  else
+    iDatum = ExecEvalExprSwitchContext(estate, econtext, &isNull, NULL);
+  if (isNull)
+    {
+      c = NULL;
+    }
+  else
+    {
+      t = DatumGetTextP(iDatum);
+      c = text2cstr(t);
+      if (t != (text *) iDatum)
+        pfree(t);
+    }
+  ExecClearTuple(econtext->ecxt_scantuple);
+  return c;
+}
+
+
+unsigned int update_index_with_tuple(IndexInfo* indexInfo,
+                                     index_info *ii,
+                                     EState *estate,
+                                     HeapTuple oldtup,
+                                     HeapTuple newtup)
+{
+  ExprContext *econtext = GetPerTupleExprContext(estate);
+  sen_rc rc = sen_success;
+  int j;
+  unsigned int num_records = 0;
+  ListCell *indexpr_item = list_head(indexInfo->ii_ExpressionsState);
+          
+  for (j = 0; j < indexInfo->ii_NumIndexAttrs; j++)
+    {
+      int keycol = indexInfo->ii_KeyAttrNumbers[j];
+      char *oldc = NULL;
+      char *newc = NULL;
+
+      if (keycol != 0)
+        { /* plain index */
+          if (oldtup)
+            oldc = get_tuple_text(econtext, oldtup, keycol, NULL);
+          if (newtup)
+            newc = get_tuple_text(econtext, newtup, keycol, NULL);
+        }
+      else 
+        { /* index on expressions */
+          if (indexpr_item == NULL)
+            elog(ERROR, "pgsenna2: wrong number of index expressions");
+          if (oldtup)
+            oldc = get_tuple_text(econtext, oldtup, 0, (ExprState *)lfirst(indexpr_item));
+          if (newtup)
+            newc = get_tuple_text(econtext, newtup, 0, (ExprState *)lfirst(indexpr_item));
+          indexpr_item = lnext(indexpr_item);
+        }
+      if (oldc || newc)
+        {
+          LOCKTAG locktag;
+          LockAcquireResult res = LOCKACQUIRE_OK;
+
+          SET_LOCKTAG_RELATION(locktag, MyDatabaseId, ii->relid);
+#if defined(POSTGRES82) || defined(POSTGRES83)
+          res = LockAcquire(&locktag, ShareUpdateExclusiveLock, 0, 0);
+#else
+          /*
+            res = LockAcquire(DEFAULT_LOCKMETHOD, &locktag, r->rd_istemp,
+            ShareUpdateExclusiveLock, 0, 0);
+          */
+#endif
+          if(res == LOCKACQUIRE_OK)
+            {
+              if (oldc)
+                {
+                  rc = sen_index_upd(ii->index,
+                                     (void *)(&(oldtup->t_self)),
+                                     oldc, strlen(oldc), NULL, 0);
+                  pfree(oldc);
+                  if (rc != sen_success)
+                    elog(ERROR, "pgsenna2: sen_index_upd failed while bulkdelete (%d)", rc);
+                }
+              if (newc)
+                {
+                  rc = sen_index_upd(ii->index,
+                                     (void *)(&(newtup->t_self)),
+                                     NULL, 0, newc, strlen(newc));
+                  elog(DEBUG1, "pgsenna2: relid: %d, newc: %s", ii->relid, newc);
+                  pfree(newc);
+                  if (rc != sen_success)
+                    elog(ERROR, "pgsenna2: sen_index_upd failed while bulkdelete (%d)", rc);
+                }
+            }
+          else
+            elog(ERROR, "pgsenna2: cannot LockAcquire while bulkdelete (%d)", res);
+
+          num_records += 1;
+#if defined(POSTGRES82) || defined(POSTGRES83)
+          LockRelease(&locktag, ShareUpdateExclusiveLock, 0);
+#else
+          /*
+            LockRelease(DEFAULT_LOCKMETHOD, &locktag, 
+            ShareUpdateExclusiveLock, 0);
+          */
+#endif
+        }
+    }
+  return num_records;
 }
